from bisect import bisect_right
from typing import List
from rich.console import RenderableType
from rich.style import Style
from rich.text import Span, Text
from textual.widget import Widget
from textual.widgets import Static
from smassh.src import master_generator, Tracker, Cursor
from smassh.src.buddy import Buddy
from smassh.src.parser import config_parser
from smassh.ui.events import ShowResults
from smassh.ui.widgets.typing.ticker import Ticker


def caret(func):
    def wrapper(space: "Space") -> Text:
        renderable: Text = func(space).copy()
        setting = config_parser.get("caret_style")
        pos = space.tracker.cursor_pos

        if setting == "off" or pos == len(space.paragraph.plain):
            return renderable

        if setting == "underline":
            rich_style = "--caret-underline"
        else:
            rich_style = "--caret-block"

        style = space.get_component_rich_style(rich_style)
        renderable.spans.append(Span(pos, pos + 1, style))

        return renderable

    return wrapper


def tab_reset(func):
    def wrapper(space: "Space", key: str) -> None:
        if key == "tab" and config_parser.get("tab_reset"):
            return space.restart()

        return func(space, key)

    return wrapper


def toggle_settings(func):
    def wrapper(space: "Space", key: str) -> None:
        config_changed = False

        if key == "ctrl+n":
            config_parser.toggle_numbers()
            config_changed = True

        elif key == "ctrl+p":
            config_parser.toggle_punctuations()
            config_changed = True

        if config_changed:
            for i in space.screen.query("Switchable"):
                i.refresh()

            return space.restart()

        return func(space, key)

    return wrapper


def cursor_buddy(func):
    def wrapper(space: "Space") -> RenderableType:
        wpm = config_parser.get("cursor_buddy_speed")
        res = func(space)

        if not wpm or not space.tracker.stats.start_time:
            return res

        elapsed = space.tracker.stats.elapsed_time
        letters_typed = Buddy.get_letters_typed(elapsed, wpm, 5)
        letters_typed = min(letters_typed, len(space.paragraph.plain) - 1)

        res_copy = res.copy()
        style = space.get_component_rich_style("--cursor-buddy")
        res_copy.spans.append(Span(letters_typed, letters_typed + 1, style))
        return res_copy

    return wrapper


def blind_mode(func):
    def wrapper(space: "Space", *args, **kwargs) -> Style:
        if config_parser.get("blind_mode") == "on":
            return space.get_component_rich_style("--blind-match")

        return func(space, *args, **kwargs)

    return wrapper


def incorrect_spaces(func):
    INCORRECT_SPACE_CHARACTER = "░"

    def wrapper(space: "Space") -> RenderableType:
        text: Text = func(space)
        incorrect_style = space.get_match_style(False)

        plain_text = text.plain

        for span in text.spans:
            if span.style == incorrect_style and plain_text[span.start] == " ":
                plain_text = (
                    plain_text[: span.start]
                    + INCORRECT_SPACE_CHARACTER
                    + plain_text[span.end :]
                )

        text.plain = plain_text
        return text

    return wrapper


class Space(Static):
    """
    Space Widget to handle keypress and display typing text
    """

    COMPONENT_CLASSES = {
        "--cursor-buddy",
        "--correct-match",
        "--incorrect-match",
        "--blind-match",
        "--caret-underline",
        "--caret-block",
    }

    def __init__(self) -> None:
        super().__init__()
        self.current_key = None
        self.reset()
        self.check_timer = self.set_interval(1, self.check_restrictions, pause=True)
        if config_parser.get("cursor_buddy_speed"):
            self.set_interval(0.1, self.refresh)

    # ---------------- UTILS -----------------

    def cursor_row(self, cursor_pos: int) -> int:
        return bisect_right(self.newlines, cursor_pos)

    def cursor_span(self, pos: int) -> Span:
        return Span(pos, pos + 1, "reverse white")

    # ----------------- RENDER ------------------
    def on_show(self) -> None:
        self.reset_newlines()

    def reset_newlines(self) -> None:
        self.newlines = master_generator.get_newlines(
            self.paragraph.plain,
            self.size.width,
        )

    def restart(self, force: bool = False) -> None:
        if force:
            generated = self.paragraph.plain
            self.paragraph = Text(generated)
            self.reset_components()
        else:
            self.reset()

    def check_restrictions(self) -> None:
        if not self.tracker.stats.start_time or self.tracker.stats.elapsed_time < 1:
            return

        if min_speed := config_parser.get("min_speed"):
            wpm = self.tracker.stats.wpm
            if wpm < min_speed:
                return self.finish_typing()

        if min_accuracy := config_parser.get("min_accuracy"):
            accuracy = self.tracker.stats.accuracy
            if accuracy < min_accuracy:
                return self.finish_typing()

        if min_burst := config_parser.get("min_burst"):
            burst = self.tracker.stats.last_word_wpm
            if burst < min_burst:
                return self.finish_typing()

    def finish_typing(self, fail: bool = True) -> None:
        self.screen.post_message(ShowResults(self.tracker.stats, fail))

    def reset(self) -> None:
        mode = config_parser.get("mode")
        if mode == "words":
            word_count = config_parser.get(f"{mode}_count")
        else:
            minutes = config_parser.get(f"{mode}_count") / 60
            word_count = round(600 * minutes)

        language = config_parser.get("language")

        generated = master_generator.generate(
            language,
            word_count,
        )

        self.paragraph = Text(generated)
        self.reset_components()

    def reset_components(self) -> None:
        self.tracker = Tracker(self.paragraph.plain)
        self.cursor = 0
        self.paragraph_spans = []

        if self.size.width:
            self.reset_newlines()
            self.screen.query_one(Ticker).reset()

        self.refresh(layout=True)

    @cursor_buddy
    @caret
    @incorrect_spaces
    def render(self) -> RenderableType:
        self.paragraph.spans = self.get_colorized()
        return self.paragraph

    @blind_mode
    def get_match_style(self, correct: bool) -> Style:
        rich_style = "correct" if correct else "incorrect"
        style = self.get_component_rich_style(f"--{rich_style}-match")
        return style

    def get_colorized(self) -> List[Span]:
        spans = []
        for index, keymatch in enumerate(self.paragraph_spans):
            if keymatch != "":
                spans.append(
                    Span(
                        index,
                        index + 1,
                        self.get_match_style(keymatch),
                    )
                )
            else:
                spans.append(
                    Span(
                        index,
                        index + 1,
                        "",
                    )
                )

        return spans

    def update_colors(self, cursor: Cursor) -> None:
        old = cursor.old
        new = cursor.new
        correct = cursor.correct

        if new < old:
            self.paragraph_spans = self.paragraph_spans[:new]
            return

        diff = new - old
        self.paragraph_spans.extend([""] * (diff - 1))

        if diff == 1:
            self.paragraph_spans.append(correct)

    # ---------------- KEYPRESS -----------------

    @toggle_settings
    @tab_reset
    def keypress(self, key: str) -> None:
        if key == "escape":
            return self.reset()

        cursor = self.tracker.keypress(key)
        if not cursor:
            return

        current_row = self.cursor_row(cursor.old)
        new_row = self.cursor_row(cursor.new)

        if current_row != new_row and isinstance(self.parent, Widget):
            if new_row > current_row:
                if current_row:
                    self.parent.scroll_down()
            else:
                if current_row != len(self.newlines) - 1:
                    self.parent.scroll_up()

        self.update_colors(cursor)
        if cursor.new == len(self.paragraph.plain):
            return self.finish_typing(fail=False)

        self.check_timer.resume()
        self.screen.query_one(Ticker).update_check.resume()
        self.refresh()



from rich.console import RenderableType
from rich.text import Text
from textual.app import ComposeResult
from textual.widget import Widget
from textual.widgets import Digits, Static

from smassh.src.parser import data_parser


class ValueLabel(Widget):
    """
    Label Widgets for values in result screen
    """

    DEFAULT_CSS = """
    ValueLabel {
        height: auto;
        width: auto;
    }
    """

    COMPONENT_CLASSES = {"--personal-best"}

    best = False

    def __init__(self, text: str, **kwargs) -> None:
        self.text = Text(text)
        super().__init__(**kwargs)

    def set_best(self, is_best: bool) -> None:
        self.best = is_best
        self.refresh()

    def render(self) -> RenderableType:
        style = self.get_component_rich_style("--personal-best")
        best_icon = Text(" 󱟜 " if self.best else "", style=style)
        return self.text + best_icon


class AutoVertical(Widget):
    DEFAULT_CSS = """
    AutoVertical {
        layout: vertical;
        margin: 1;
        height: auto;
        width: auto;
    }
    """


class Value(Digits, Static):
    """
    Widget to show result values in larger font size
    """

    DEFAULT_CSS = """
    Value {
        margin: 0 2;
        content-align: right middle;
        height: auto;
        width: auto;
    }
    """


class ValueContainer(Static):
    """
    Container widget that holds all the Values
    """

    DEFAULT_CSS = """
    ValueContainer {
        layout: horizontal;
        margin: 0 2;
        height: auto;
        width: 100%;
        align: center middle;
    }
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.wpm = Value()
        self.accuracy = Value()

    def update_stats(self, stats) -> None:
        wpm_label = self.query_one("#wpm_label", expect_type=ValueLabel)
        wpm_label.set_best(data_parser.is_highest_wpm(stats.wpm))

        acc_label = self.query_one("#acc_label", expect_type=ValueLabel)
        acc_label.set_best(data_parser.is_highest_accuracy(stats.accuracy))

        self.wpm.update(str(stats.wpm))
        self.accuracy.update(str(stats.accuracy))

    def compose(self) -> ComposeResult:
        with AutoVertical():
            yield ValueLabel("WPM", id="wpm_label")
            yield self.wpm

        with AutoVertical():
            yield ValueLabel("ACC", id="acc_label")
            yield self.accuracy



from pathlib import Path
from typing import Any, Dict, List
from .parser import Parser
from smassh.src.stats_tracker import StatsTracker
from smassh.src.parser.config_parser import config_parser
import appdirs


class DataParser(Parser):
    """
    Inherited from `Parser` class to manage user data
    """

    config_path = Path(appdirs.user_data_dir("smassh"))
    lang_path = config_path / "languages"
    DEFAULT_CONFIG = dict(data=[])

    def __init__(self) -> None:
        super().__init__()

        english_path = self.lang_path / "english.json"

        if not self.lang_path.is_dir():
            self.lang_path.mkdir(parents=True, exist_ok=True)

        if not english_path.exists():
            from smassh.src.plugins.add_language import AddLanguage

            AddLanguage(silent=True).add("english")

    def generate_report(self, stats: StatsTracker) -> Dict[str, Any]:
        mode = config_parser.get("mode")
        count = config_parser.get(f"{mode}_count")
        start = stats.start_time or 0
        end = stats.end_time or 0
        elapsed = end - start

        return dict(
            mode=mode,
            count=count,
            start_time=stats.start_time,
            elapsed_time=elapsed,
            wpm=stats.wpm,
            raw_wpm=stats.raw_wpm,
            accuracy=stats.accuracy,
        )

    def add_stats(self, stats: StatsTracker, failed: bool) -> None:
        report = self.generate_report(stats) | dict(failed=failed)
        self.get("data").append(report)
        self.save()

    def current_mode_tests(self) -> List[Dict]:
        mode = config_parser.get("mode")
        count = config_parser.get(f"{mode}_count")

        def same_mode(test: Dict):
            return test["mode"] == mode and count == test["count"]

        return list(filter(same_mode, self.get("data")))

    def hightest_wpm(self) -> int:
        tests = self.current_mode_tests()
        if not tests:
            return 0

        return max(tests, key=lambda x: x["wpm"])["wpm"]

    def hightest_accuracy(self) -> int:
        tests = self.current_mode_tests()
        if not tests:
            return 0

        return max(tests, key=lambda x: x["accuracy"])["accuracy"]

    def is_highest_wpm(self, wpm: int) -> bool:
        return wpm > self.hightest_wpm()

    def is_highest_accuracy(self, accuracy: int) -> bool:
        return accuracy > self.hightest_accuracy()


data_parser = DataParser()



from json import load, dump
from pathlib import Path
from typing import Any, Dict


def combine_into(d: dict, to: dict) -> None:
    for k, v in d.items():
        if isinstance(v, dict):
            combine_into(v, to.setdefault(k, {}))
        else:
            to[k] = v


class Parser:
    """
    The base class for ConfigParser and DataParser classes
    """

    _file_name: str = "smassh"
    config_path: Path
    DEFAULT_CONFIG: Dict[str, Any]

    def __init__(self) -> None:
        self.config = self.DEFAULT_CONFIG.copy()
        if not Path.is_file(self.full_path):
            self._create_user_config()
        else:
            self.update(self.read_from_file())

    @property
    def file_name(self) -> str:
        return self._file_name + ".json"

    @property
    def full_path(self) -> Path:
        return self.config_path.joinpath(self.file_name)

    def set(self, key: str, value: Any) -> None:
        self.config[key] = value
        self.save()

    def update(self, data: Dict[str, Any]) -> None:
        combine_into(data, self.config)

    def save(self) -> None:
        with open(self.full_path, "w") as fp:
            dump(self.config, fp)

    def reset(self):
        self.config = self.DEFAULT_CONFIG
        self.save()

    def _create_user_config(self) -> None:
        """
        Creates a new config
        """

        self.config_path.mkdir(parents=True, exist_ok=True)
        self.save()

    def get(self, data: str) -> Any:
        return self.config.get(data)

    def read_from_file(self) -> Dict[str, Any]:
        with open(self.full_path, "r") as fp:
            return load(fp)



class Buddy:
    """
    Buddy class to calculate the location of cursor buddy
    """

    @classmethod
    def get_letters_typed(cls, elapsed: float, wpm: int, parity_constant: float) -> int:
        words_typed = elapsed * wpm
        letters_typed = words_typed / parity_constant
        return round(letters_typed)



from typing import List
from textual.app import ComposeResult
from textual import events
from textual.widget import Widget
from textual.widgets import Static
from smassh.ui.widgets import menu
from smassh.ui.widgets import BaseWindow
from smassh.ui.widgets.settings.settings_options import Setting
from smassh.ui.widgets.settings import (
    SettingStrip,
    SettingStripItem,
    SettingSeparator,
    Bracket,
)


class SettingGrid(Widget):
    DEFAULT_CSS = """
    SettingGrid {
        layout: grid;
        grid-size: 1 2;
        grid-rows: 2 1fr;
    }
    """


class SettingsContainer(Widget):
    DEFAULT_CSS = """
    SettingsContainer {
        width: 1fr;
        height: 1fr;
        layout: vertical;
        overflow: auto auto;
    }
    """


class SettingsScreen(BaseWindow):
    """
    Setting Screen to show all the tweakables!
    """

    DEFAULT_CSS = """
    SettingsScreen {
        padding: 1;
    }
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.current_option = 0
        self.sections: List[SettingStripItem] = []
        self.settings: List[Setting] = []

        for section, settings in menu.items():
            self.sections.append(SettingStripItem(section))
            self.settings.extend(settings)

    @property
    def current_setting(self) -> Setting:
        return self.settings[self.current_option]

    def get_section(self, setting: Setting) -> str:
        for section, settings in menu.items():
            if setting in settings:
                return section

        raise ValueError(f"Setting {setting} not found in menu")

    def compose(self) -> ComposeResult:

        with SettingGrid():
            with SettingStrip():
                yield Bracket("left")
                yield from self.sections
                yield Bracket("right")

            with SettingsContainer():
                for section, settings in menu.items():
                    with Static() as container:
                        yield SettingSeparator(section)
                        settings[0].set_section_widget(container)
                        yield settings[0]

                    for setting in settings[1:]:
                        yield setting

        self.update_highlight()

    def update_highlight_strip(self, section_name: str) -> None:
        for section in self.sections:
            section.set_class(section.section == section_name, "enabled")

    def update_highlight(self) -> None:
        """
        Automatically update highlights based on current option
        """

        for index, setting in enumerate(self.settings):
            is_current = index == self.current_option
            setting.set_class(is_current, "selected")
            if is_current:
                self.update_highlight_strip(self.get_section(setting))

        section = self.get_section(self.current_setting)
        if self.current_setting == menu[section][0]:
            self.current_setting.section_widget.scroll_visible()
        else:
            self.current_setting.scroll_visible()

        self.refresh()

    async def handle_key(self, event: events.Key) -> None:
        key = event.key
        n = len(self.settings)

        await super().handle_key(event)

        if key in ["down", "j"]:
            if self.current_option < n - 1:
                self.current_option += 1
                self.update_highlight()

        elif key in ["up", "k"]:
            if self.current_option > 0:
                self.current_option -= 1
                self.update_highlight()

        elif key == "tab":
            self.current_setting.select_next()

        elif key == "shift+tab":
            self.current_setting.select_prev()



from rich.console import RenderableType
from textual.widget import Widget
from textual.widgets import Label
from smassh.ui.widgets.settings.settings_options import menu


class SettingStripItem(Label):
    DEFAULT_CSS = """
    SettingStripItem {
        color: white;
        padding: 0 2;
        height: 1;
    }
    """

    def __init__(self, section: str, *args, **kwargs) -> None:
        super().__init__(section, *args, **kwargs)
        self.section = section

    def render(self) -> RenderableType:
        return self.section.replace("_", " ")

    def on_click(self) -> None:
        from smassh.ui.screens import SettingsScreen

        settings_screen = self.screen.query_one(SettingsScreen)
        settings = settings_screen.settings
        first_setting = menu[self.section][0]
        settings_screen.current_option = settings.index(first_setting)
        settings_screen.update_highlight()


class SettingStrip(Widget):
    DEFAULT_CSS = """
    SettingStrip {
        height: 2;
        layout: horizontal;
        width: 100%;
        align: center top;
    }
    """



from textual.app import ComposeResult
from textual.widget import Widget
from .palette import Palette


class LanguagePalette(Palette):
    screen_name = "language"
    config_name = "language"
    icon = ""


class ThemePalette(Palette):
    screen_name = "theme"
    config_name = "theme"
    icon = ""


class PaletteOptions(Widget):
    """
    Typing Space Widget which shows current enable palette options
    """

    DEFAULT_CSS = """
    PaletteOptions {
        layout: horizontal;
        align: center middle;
        height: 1;
        column-span: 3;
    }
    """

    def compose(self) -> ComposeResult:
        yield LanguagePalette()
        yield ThemePalette()



from textual import on, work
from textual.containers import Vertical
from textual.app import ComposeResult
from textual.binding import Binding
from textual.screen import Screen
from textual.widgets import Label
from smassh.ui.widgets import (
    PaletteList,
    PaletteInput,
    LanguagePaletteList,
    ThemePaletteList,
)


class PaletteMenu(Vertical):
    """
    A menu to show all the avaialble options for the setting
    """

    DEFAULT_CSS = """
    PaletteMenu {
        width: 60%;
        height: 50%;
        layout: grid;
        grid-size: 1 3;
        grid-rows: 2 3 1fr;
    }
    """


class PaletteHeader(Label):
    """
    Header widget to show setting name of palette menu
    """

    DEFAULT_CSS = """
    PaletteHeader {
        content-align: center middle;
        width: 100%;
    }
    """


class PaletteScreen(Screen):
    """
    Screen for palette with dark mask over main window
    """

    DEFAULT_CSS = """
    PaletteScreen {
        layout: vertical;
        align: center middle;
    }
    """
    BINDINGS = [
        Binding("escape", "app.pop_screen"),
        Binding("down", "next_option"),
        Binding("up", "prev_option"),
    ]

    palette_list: PaletteList
    palette_icon: str
    palette_header: str

    @on(PaletteInput.Changed)
    @work(thread=True, exclusive=True)
    def update_palette_list(self, event: PaletteInput.Changed) -> None:
        """
        Function to filter changes based on Input field in Palette

        Args:
            event (PaletteInput.Changed): Change Event
        """

        event.stop()
        text = self.query_one(PaletteInput).value
        self.query_one(PaletteList).apply_filter(text)

    @on(PaletteList.OptionHighlighted)
    def preview_palette_option(self, event: PaletteList.OptionHighlighted) -> None:
        """
        Function to preview change in smassh

        Args:
            event (PaletteList.OptionHighlighted): Change Event
        """

        palette_list = self.query_one(PaletteList)
        prompt = str(event.option.prompt)
        prompt = prompt.replace(" ", "_")
        self.post_message(palette_list._highlight_event(prompt))

    def action_next_option(self) -> None:
        self.query_one(PaletteList).action_cursor_down()

    def action_prev_option(self) -> None:
        self.query_one(PaletteList).action_cursor_up()

    def compose(self) -> ComposeResult:
        with PaletteMenu():
            yield PaletteHeader(self.palette_header)
            yield PaletteInput()
            yield self.palette_list


class LanguagePaletteScreen(PaletteScreen):
    palette_list = LanguagePaletteList()
    palette_icon = "  "
    palette_header = f"{palette_icon} language"


class ThemePaletteScreen(PaletteScreen):
    palette_list = ThemePaletteList()
    palette_icon = "  "
    palette_header = f"{palette_icon} themes"



from functools import cache
from typing import List, Type
from textual.message import Message
from textual.widgets import OptionList
from smassh.src import config_parser


class PaletteOptionHighlighted(Message):
    """
    Highlight Event for Palette when option is changed
    """

    def __init__(self, value: str) -> None:
        super().__init__()
        self.value = value


class ApplyLanguage(PaletteOptionHighlighted):
    pass


class ApplyTheme(PaletteOptionHighlighted):
    pass


class PaletteList(OptionList, can_focus=False):
    """
    List Widget to list all the available options
    """

    DEFAULT_CSS = """
    PaletteList {
        border: none;
        scrollbar-size: 0 1;

        .option-list--option {
            padding: 0 1;
        }

    }
    """

    _filter: str = ""
    _highlight_event: Type[PaletteOptionHighlighted]

    def _get_options(self) -> List[str]:
        raise NotImplementedError

    @cache
    def get_options(self) -> List[str]:
        options = self._get_options()
        return sorted([option.replace("_", " ") for option in options])

    def get_current(self) -> str:
        raise NotImplementedError

    def apply_filter(self, filter_text: str) -> None:

        def valid_option(option_text: str) -> bool:
            if not filter_text:
                return True

            return filter_text in option_text

        self.clear_options()
        self._filter = filter_text
        valid_options = filter(valid_option, self.get_options())
        self.add_options(valid_options)

    async def on_mount(self, _) -> None:
        self.apply_filter("")
        options = sorted(self._get_options())
        index = options.index(self.get_current())
        self.highlighted = index


class LanguagePaletteList(PaletteList):
    _highlight_event = ApplyLanguage

    def _get_options(self) -> List[str]:
        return config_parser.configured_languages

    def get_current(self) -> str:
        return config_parser.get("language")


class ThemePaletteList(PaletteList):
    _highlight_event = ApplyTheme

    def _get_options(self) -> List[str]:
        return config_parser.configured_themes

    def get_current(self) -> None:
        return config_parser.get("theme")



from rich.console import RenderableType
from rich.text import Text
from textual.widget import Widget


class SettingSeparator(Widget):
    DEFAULT_CSS = """
    SettingSeparator {
        height: auto;
        width: 100%;
        text-style: bold italic;
        margin: 1 0;
    }
    """

    COMPONENT_CLASSES = {"--primary-bg", "--danger-bg"}

    def __init__(self, section: str, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.section = section

    def render(self) -> RenderableType:
        classname = "danger" if "danger" in self.section else "primary"
        style = self.get_component_rich_style(f"--{classname}-bg")

        text = Text(self.section.upper().replace("_", " "), style=style)
        text.pad(1)
        return text



import re
import os
import sys


def convert_css(input_file, output_file) -> None:
    with open(input_file, "r") as f:
        css_content = f.read()

    pattern = re.compile(r"--([\w-]+):\s*([^;]+);")
    matches = pattern.findall(css_content)

    new_tcss_content = ""
    for custom_property, color_value in matches:
        new_tcss_content += f"${custom_property}: {color_value};\n"

    with open(output_file, "w") as f:
        f.write(new_tcss_content)


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("Usage: python script.py input.css")
        sys.exit(1)

    input_file = sys.argv[1]
    output_file = os.path.splitext(input_file)[0] + ".tcss"

    convert_css(input_file, output_file)



from textual.app import ComposeResult
from textual.containers import Container
from textual.widgets import Static
from smassh.src.stats_tracker import StatsTracker
from smassh.ui.widgets import BaseWindow
from smassh.ui.widgets.typing.space import Space
from smassh.ui.widgets.result import ValueContainer, Value, ResultStrip
from smassh.ui.widgets.typing.ticker import Ticker


class AutoVertical(Static):
    DEFAULT_CSS = """
    AutoVertical {
        margin: 1;
        height: auto;
        width: auto;
        background: red;
    }
    """


class AutoHorizontal(Static):
    DEFAULT_CSS = """
    AutoVertical {
        layout: horizontal;
        margin: 1;
        height: auto;
        width: auto;
    }
    """


class ResultScreen(BaseWindow):
    """
    This screen will show the result of the typing test.
    E.g. Typing Chart, Accuracy, WPM etc.
    """

    DEFAULT_CSS = """
    ResultScreen {
        layout: grid;
        grid-size: 1 4;
        grid-rows: 1fr auto 1 1fr;
        align: center middle;
    }
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.wpm = Value()
        self.accuracy = Value()

    def compose(self) -> ComposeResult:
        yield Container()
        yield ValueContainer()
        yield ResultStrip()
        yield Container()

    def set_results(self, stats: StatsTracker) -> None:
        self.stats = stats
        self.query_one(ValueContainer).update_stats(stats)

    def on_hide(self) -> None:
        space = self.screen.query_one(Space)
        if space.paragraph.spans:
            self.screen.query_one(Ticker).reset()
            space.restart()



from .palette_list import PaletteList, LanguagePaletteList, ThemePaletteList
from .palette_options import PaletteOptions, LanguagePalette, ThemePalette
from .palette_input import PaletteInput

__all__ = [
    "PaletteList",
    "LanguagePaletteList",
    "ThemePaletteList",
    "PaletteOptions",
    "LanguagePalette",
    "ThemePalette",
    "PaletteInput",
]



import click

PKG_VERSION = "3.1.3"


@click.group(
    context_settings={"help_option_names": ["-h", "--help"]},
    invoke_without_command=True,
)
@click.option(
    "--version",
    "-v",
    is_flag=True,
    help="Show version and exit.",
)
@click.pass_context
def main(ctx, version: bool) -> None:
    if version:
        return print(f"smassh - v{PKG_VERSION}")

    if ctx.invoked_subcommand is None:
        from smassh.ui.tui import Smassh

        Smassh().run()


@main.command(help="Add a language to smassh")
@click.argument("name")
def add(name: str) -> None:
    from smassh.src.plugins.add_language import AddLanguage

    AddLanguage().add(name)


if __name__ == "__main__":
    main()



from typing import Callable, List, Optional
from rich.console import RenderableType
from rich.text import Text
from textual.app import ComposeResult
from textual.widget import Widget
from textual.widgets import Label
from smassh.src import config_parser
from smassh.ui.widgets.typing.space import Space


class BaseOption(Widget):
    """
    Base Option Widget to extent from for setting
    """

    DEFAULT_CSS = """
    BaseOption {
        align: center middle;
        content-align: center middle;
        padding: 0 1;
    }
    """

    def __init__(self, setting_name: str, callback: Optional[Callable] = None) -> None:
        super().__init__(id=f"option-{setting_name}")
        self.setting_name = setting_name
        self._callback = callback

    def on_mount(self) -> None:
        self.load_current_setting()

    def callback(self) -> None:
        if self._callback:
            return self._callback()

        return self.app.query_one(Space).reset()

    @property
    def value(self) -> str:
        raise NotImplementedError

    def highlight(self) -> None:
        pass

    def lowlight(self) -> None:
        pass

    def _select_next_option(self) -> None:
        pass

    def _select_prev_option(self) -> None:
        pass

    def select_next_option(self) -> None:
        self._select_next_option()
        self.save()

    def select_prev_option(self) -> None:
        self._select_prev_option()
        self.save()

    def save(self) -> None:
        config_parser.set(self.setting_name, self.value)
        self.callback()

    def load_current_setting(self) -> None:
        raise NotImplementedError


class ConfirmLabel(Label):
    DEFAULT_CSS = """
    ConfirmLabel {
        padding: 0 1;
    }
    """

    def on_click(self):
        if self.parent and isinstance(self.parent, Confirm):
            self.parent.select_next_option()


class Confirm(BaseOption):

    def __init__(
        self,
        setting_name: str,
        message: str,
        callback: Callable = lambda: None,
    ) -> None:
        super().__init__(setting_name, callback)
        self.message = message

    def load_current_setting(self) -> None:
        pass

    def select_prev_option(self) -> None:
        pass

    def post_confirm(self, ok: bool) -> None:
        if ok:
            self.save()

    def save(self) -> None:
        self.callback()

    def select_next_option(self) -> None:
        self.app.push_screen("confirm", self.post_confirm)

    def compose(self) -> ComposeResult:
        yield ConfirmLabel(self.message)


class OptionItem(Widget):
    """
    Widget for sigle option in `Option` widget
    """

    DEFAULT_CSS = """
    OptionItem {
        height: 1;
        width: auto;
        padding: 0 1;
    }
    """

    def __init__(self, value: str) -> None:
        super().__init__()
        self.value = value

    def on_click(self) -> None:
        parent = self.parent
        if isinstance(parent, Option):
            self_index = parent.options.index(self)
            parent._value = self_index
            parent.update_highlight()
            parent.save()

    def render(self) -> RenderableType:
        return self.value


class Option(BaseOption):
    """
    Option Widget for choosing between multiple options
    """

    def __init__(
        self,
        setting_name: str,
        options: List[str],
        callback: Optional[Callable] = None,
    ) -> None:
        self.options = [OptionItem(option) for option in options]
        self._value = 0
        super().__init__(setting_name, callback)

    def load_current_setting(self) -> None:
        setting = config_parser.get(self.setting_name)
        if isinstance(setting, bool):
            setting = ["off", "on"][setting]

        option = [i for i in self.options if i.value == setting][0]
        self._value = self.options.index(option)
        self.update_highlight()

    @property
    def value(self) -> str:
        return self.options[self._value].value

    def update_highlight(self) -> None:
        for i, option in enumerate(self.options):
            option.set_class(i == self._value, "selected")

    def _select_next_option(self) -> None:
        n = len(self.options)
        self._value = (self._value + 1) % n
        self.update_highlight()

    def _select_prev_option(self) -> None:
        n = len(self.options)
        self._value = (self._value - 1 + n) % n
        self.update_highlight()

    def compose(self) -> ComposeResult:
        for option in self.options:
            yield option


class NumberScroll(BaseOption):
    """
    Widget for setting numeric value for a setting
    """

    COMPONENT_CLASSES = {"scroll--background"}

    def __init__(
        self,
        setting_name: str,
        callback: Optional[Callable] = None,
    ) -> None:
        super().__init__(setting_name, callback)
        self._value = 0

    @property
    def value(self) -> int:
        return self._value

    def load_current_setting(self) -> None:
        value = config_parser.get(self.setting_name)
        self._value = value

    def _select_next_option(self) -> None:
        self._value = min(self._value + 1, 100)
        self.refresh()

    def _select_prev_option(self) -> None:
        self._value = max(self._value - 1, 0)
        self.refresh()

    def render(self) -> RenderableType:
        style = self.get_component_rich_style("scroll--background")
        text = Text(str(self.value), style=style)
        text.pad(2)
        return text



from textual.widgets import Input


class PaletteInput(Input):
    """
    Input Widget for Palette Menu to filter results
    """

    DEFAULT_CSS = """
    PaletteInput {
        border: none;
        content-align: left middle;
        padding: 0 1;
    }

    PaletteInput:focus {
        border: none;
        content-align: left middle;
        padding: 0 1;
    }
    """



from rich.console import RenderableType
from textual.app import ComposeResult, events
from textual.widget import Widget
from smassh.ui.widgets import (
    BaseWindow,
    TypingConfigStrip,
    PaletteOptions,
    Space,
    Ticker,
)
from textual.containers import VerticalScroll
from smassh.ui.events import SetScreen


class TypingScroll(VerticalScroll, can_focus=False):
    pass


class Pad(Widget):
    """
    Pad widget for empty spaces
    """

    DEFAULT_CSS = """
    Pad.cspan3 {
        column-span: 3;
    }
    """

    def render(self) -> RenderableType:
        return ""


class TypingSpace(Widget):
    """
    Widget that holds all the test paragraph and time ticker
    """

    DEFAULT_CSS = """
    TypingSpace {
        layout: grid;
        grid-size: 3 6;
        grid-columns: 1fr 4fr 1fr;
        grid-rows: 1 1fr 1 3 1fr 1;
        margin: 1;
        height: 100%;
        width: 100%;
    }

    VerticalScroll {
        scrollbar-size: 0 0;
    }

    """

    def config_strip(self) -> ComposeResult:
        yield TypingConfigStrip()

    def counter(self) -> ComposeResult:
        yield Pad()
        yield Ticker()
        yield Pad()

    def space(self) -> ComposeResult:
        yield Pad()
        with TypingScroll():
            yield Space()
        yield Pad()

    def compose(self) -> ComposeResult:
        yield from self.config_strip()
        yield Pad(classes="cspan3")
        yield from self.counter()
        yield from self.space()
        yield Pad(classes="cspan3")
        yield PaletteOptions()

    def keypress(self, key: str):
        if key == "ctrl+s":
            return self.screen.post_message(SetScreen("settings"))

        if key == "ctrl+l":
            return self.app.push_screen("language")

        if key == "ctrl+t":
            return self.app.push_screen("theme")

        self.query_one(Space).keypress(key)


class TypingScreen(BaseWindow):
    """
    Screen Widget for typing area!
    """

    def compose(self) -> ComposeResult:
        yield TypingSpace()

    async def handle_key(self, event: events.Key):
        if not self.visible:
            return

        event.stop()
        key = event.character if event.is_printable and event.character else event.key
        self.query_one(TypingSpace).keypress(key)



from rich.console import RenderableType
from rich.text import Text
from textual.app import ComposeResult
from textual.widget import Widget
from smassh.ui.widgets import BaseWindow

DESCRIPTION = """
Smassh is a TUI based typing test application inspired by MonkeyType
-- A very popular online web-based typing application

All the other TUI based applications I found were not graphically appealing
and then I discovered textual which made it possible to create this application.

Smassh tries to be a full fledged typing test experience but not missing
out on looks and feel!

A lot of work is still left to be done and I'll be more than happy to accept
any ideas and contributions :D

Thanks for checking out this project!
"""

STAR = """
[@click=app.star]Star this project on GitHub![/]
"""

OUTRO = """
Made with ❤️  by kraanzu
"""


class Section(Widget):
    """
    Section Widget for each section of About Menu
    """

    DEFAULT_CSS = """
    Section {
        height: auto;
        content-align: center middle;
    }
    """

    def __init__(self, renderable: str) -> None:
        super().__init__()
        self.renderable = renderable

    def render(self) -> RenderableType:
        return Text.from_markup(self.renderable)


class AboutDescription(Section):
    def __init__(self) -> None:
        super().__init__(DESCRIPTION)


class Star(Section):
    def __init__(self) -> None:
        super().__init__(STAR)


class AboutOutro(Section):
    DEFAULT_CSS = """
    AboutOutro {
        content-align: center bottom;
        height: 100%;
    }
    """

    def __init__(self) -> None:
        super().__init__(OUTRO)


class AboutScreen(BaseWindow):
    """
    About screen to show info about the project
    """

    DEFAULT_CSS = """
    AboutScreen {
        layout: grid;
        grid-size: 1 3;
        grid-rows: auto auto 1fr;
    }
    """

    def compose(self) -> ComposeResult:
        yield AboutDescription()
        yield Star()
        yield AboutOutro()



from appdirs import user_cache_dir
from pathlib import Path

TARGET_FOLDER = Path(user_cache_dir("smassh"))
TARGET_FILE = TARGET_FOLDER / "styles.tcss"


def write_css_file(theme_css: str, base_css: str) -> None:
    if not TARGET_FILE.exists():
        TARGET_FOLDER.mkdir(parents=True, exist_ok=True)

    with open(TARGET_FILE, "w") as target:
        target.write(theme_css)
        target.write(base_css)


def generate_theme_file(theme: str) -> None:
    """
    Theme generator which merges theme and base CSS files

    Args:
        theme (str): theme name
    """

    css_folder = Path.absolute(Path(__file__).parent.parent) / "ui" / "css"
    themes_folder = css_folder / "themes"

    base_path = css_folder / "base.tcss"
    theme_path = themes_folder / f"{theme}.tcss"

    with open(theme_path, "r") as theme_file:
        theme_css = theme_file.read()

    with open(base_path, "r") as base_file:
        base_css = base_file.read()

    write_css_file(theme_css, base_css)



from smassh.src.parser import config_parser
from .option import BaseOption, Confirm


class ResetConfig(Confirm):
    def __init__(self) -> None:

        def callback():
            config_parser.reset()
            for setting in self.screen.query(BaseOption):
                setting.load_current_setting()
                setting.refresh()

        super().__init__("rest_config", "Reset", callback)



from time import time
from dataclasses import dataclass
from enum import Enum
from typing import List


class Match(Enum):
    """
    Match enum class for character match
    """

    MATCH = 1
    MISMATCH = 2
    BACKSPACE = 3
    SKIPPED = 4


@dataclass
class CheckPoint:
    """
    Checkpoint class to maintain record of position and elapsed time at that point
    """

    letter: str
    position: int
    correct: Match

    def add_elapsed(self, elapsed) -> None:
        self.elapsed = elapsed


class StatsTracker:
    """
    Tracker class to calculate stats while typing
    """

    def __init__(self) -> None:
        self.reset()

    def get_checkpoints_last_word(self) -> List[CheckPoint]:
        checkpoints = self.checkpoints.copy()
        word_checkpoints = []

        while checkpoints and checkpoints[-1].letter == " ":
            checkpoints.pop()

        while checkpoints and checkpoints[-1].letter != " ":
            word_checkpoints.append(checkpoints.pop())

        return list(reversed(word_checkpoints))

    @property
    def elapsed_time(self) -> float:
        if not self.start_time:
            raise ValueError("Start time not set")

        if self.end_time:
            return self.end_time - self.start_time

        return time() - self.start_time

    @property
    def word_count(self) -> int:
        if not self.checkpoints:
            return 0

        return sum(checkpoint.letter == " " for checkpoint in self.checkpoints)

    @property
    def last_word_accuracy(self) -> int:
        correct = 0
        incorrect = 0

        checkpoints = self.get_checkpoints_last_word()

        if not checkpoints:
            raise ValueError("No checkpoints")

        for i in checkpoints:
            correct += i.correct == Match.MATCH
            incorrect += i.correct == Match.MISMATCH

        return round((correct / (correct + incorrect)) * 100)

    @property
    def last_word_wpm(self) -> int:
        checkpoints = self.get_checkpoints_last_word()

        if not checkpoints:
            raise ValueError("No checkpoints")

        start = checkpoints[0].elapsed
        stop = checkpoints[-1].elapsed
        elapsed = stop - start

        if elapsed == 0:
            raise ValueError("Elapsed time is 0")

        raw = 60 / elapsed
        return round(self.last_word_accuracy * raw / 100)

    @property
    def raw_wpm(self) -> int:
        # TODO: Better formula because this is not accurate

        time_taken = self.elapsed_time / 60
        return round(self.word_count / time_taken)

    @property
    def accuracy(self) -> int:
        total_typed = self.correct + self.incorrect

        if total_typed == 0:
            return 0

        accuracy = (self.correct / total_typed) * 100

        return round(accuracy)

    @property
    def wpm(self) -> int:
        return round(self.raw_wpm * (self.accuracy / 100))

    @property
    def correct(self) -> int:
        return sum(checkpoint.correct == Match.MATCH for checkpoint in self.checkpoints)

    @property
    def incorrect(self) -> int:
        return sum(
            checkpoint.correct == Match.MISMATCH for checkpoint in self.checkpoints
        )

    @property
    def missed(self) -> int:
        return sum(
            checkpoint.correct == Match.SKIPPED for checkpoint in self.checkpoints
        )

    # ---------------------------------------

    def reset(self) -> None:
        self.start_time = None
        self.end_time = None
        self.checkpoints: List[CheckPoint] = []

    def finish(self) -> None:
        self.end_time = time()

    def add_checkpoint(self, checkpoint: CheckPoint) -> None:
        if not self.start_time:
            self.start_time = time()

        if checkpoint.correct == Match.BACKSPACE:
            diff = self.checkpoints[-1].position - checkpoint.position
            self.checkpoints = self.checkpoints[:-diff]
            return

        elapsed = time() - self.start_time
        checkpoint.add_elapsed(elapsed)

        self.checkpoints.append(checkpoint)



from textual.message import Message
from smassh.src import StatsTracker


class SetScreen(Message):
    """
    Emitted to change the screen content
    """

    def __init__(self, screen_name: str) -> None:
        super().__init__()
        self.screen_name = screen_name


class ShowResults(Message):
    """
    Emitted when the typing is finished
    """

    def __init__(self, stats: StatsTracker, failed: bool = False) -> None:
        super().__init__()
        self.stats = stats
        self.failed = failed
